"""Self-Delimiting Numeric Values (RFC 6256)"""


def sdnv_encode(value):
    """Returns the SDNV-encoded byte string representing value

    Args:
        value (int): Non-negative integer that should be encoded
    Returns:
        bytes: SDNV-encoded number
    Raises:
        ValueError: If value not an integer or negative
    """
    if value < 0 or not isinstance(value, int):
        raise ValueError("Only non-negative integers can be SDNV-encoded")

    # Special case: zero whould produce an empty byte string
    if value == 0:
        return b"\x00"

    result = bytearray()
    while value != 0:
        result.append((value & 0x7f) | 0x80)
        value >>= 7

    # Clear MSB in the first byte that was set by the XOR with 0x80
    result[0] &= 0x7f

    # Network byte order
    return bytes(reversed(result))


def sdnv_decode(buffer):
    """Decodes a byte string (or any iterable over bytes) assumed to be a an
    SDNV and returns the non-negative integer representing the numeric value

    Args:
        buffer (bytes): Encoded SDNV
    Returns:
        int: Decoded non-negative integer
    Raises:
        ValueError: If the buffer contains insufficient bytes (not the complete
        SDNV)
    """
    n = 0
    for i, byte in enumerate(buffer, 1):
        n = (n << 7) | (byte & 0x7f)

        if byte >> 7 == 0:
            return n, i

    raise ValueError("Insufficient bytes")


def test_sdnv_encode():
    assert sdnv_encode(0xabc) == b"\x95\x3c"
    assert sdnv_encode(0x4234) == b"\x81\x84\x34"


def test_sdnv_decode():
    import pytest

    assert sdnv_decode(b"\xa4\x34") == (0x1234, 2)
    assert sdnv_decode(b"\xa4\x34\x00\x00") == (0x1234, 2)
    assert sdnv_decode(b"\x7f") == (0x7f, 1)

    with pytest.raises(ValueError):
        assert sdnv_decode(b"\xa4")



import socket
import logging

import cbor

from .helpers import sock_recv_raw

logger = logging.getLogger(__name__)


class MTCPSocket(object):
    """Minimal TCP convergence layer socket wrapper."""

    def __init__(self, sock, timeout=None):
        self.sock = sock
        self.timeout = timeout

    def __enter__(self):
        self.connect()
        return self

    def __exit__(self, *args):
        self.disconnect()

    def connect(self):
        """Establish an MTCP connection to a remote peer."""
        # Expect a connected socket by default.
        pass

    def disconnect(self):
        """Close an MTCP connection."""
        logger.debug(
            "MTCP: Terminate connection to %s:%d",
            *self.sock.getpeername(),
        )
        self.sock.shutdown(socket.SHUT_RDWR)
        self.sock.close()

    def send_bundle(self, bundle):
        """Send a single bundle to the remote peer.

        Args:
            bundle (bytes): Serialized bundle
        """
        return self.sock.sendall(cbor.dumps(bundle))

    def recv_bundle(self):
        """Receives a single bundle from the remote peer.

        Returns:
            bundle (bytes): Serialized bundle
        """
        # Receive first byte of CBOR byte string header
        length_type_raw = sock_recv_raw(self.sock, 1, self.timeout)
        additional_information = length_type_raw[0] & ((1 << 5) - 1)
        assert additional_information <= 27
        # Determine the count of following bytes
        if additional_information >= 24:
            additional_bytes = (1 << (additional_information - 24))
            length_type_raw += sock_recv_raw(
                self.sock, additional_bytes, self.timeout
            )
        # Convert CBOR byte string header to uint and decode length
        length_type_cbor_uint = (
            bytes([(length_type_raw[0] & ~0x40)]) +
            length_type_raw[1:]
        )
        byte_string_length = cbor.loads(length_type_cbor_uint)
        # Receive following byte string which is our payload
        data = sock_recv_raw(
            self.sock,
            byte_string_length,
            self.timeout,
        )
        return data


class MTCPConnection(MTCPSocket):
    """Minimal TCP convergence layer connection to a remote DTN node.

    The connection supports the context manager protocol to ensure a proper
    shutdown of the connection in case of any error:

    .. code:: python

        from pyd3tn.mtcp import MTCPConnection

        with MTCPConnection('127.0.0.1', 4224) as conn:
            conn.send(bundle)

    Args:
        host (str): Host to connect to
        port (int): TCP port to connect to
        timeout (float): Timeout for receiving bundles (None for infinite)
    """

    def __init__(self, host, port, timeout=None):
        self.host = host
        self.port = port
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        super().__init__(sock, timeout)

    def connect(self):
        """Establish an MTCP connection to a remote peer."""
        self.sock.connect((self.host, self.port))
        logger.debug("MTCP: Connected to %s:%d", self.host, self.port)



import select
from time import time
from datetime import datetime, timezone

UNIX_EPOCH = datetime(1970, 1, 1, tzinfo=timezone.utc)
DTN_EPOCH = datetime(2000, 1, 1, tzinfo=timezone.utc)
CCSDS_EPOCH = datetime(1958, 1, 1, tzinfo=timezone.utc)

UNIX_TO_DTN_OFFSET = (DTN_EPOCH - UNIX_EPOCH).total_seconds()
assert UNIX_TO_DTN_OFFSET == 946684800

CCSDS_TO_UNIX_OFFSET = (UNIX_EPOCH - CCSDS_EPOCH).total_seconds()
CCSDS_TO_DTN_OFFSET = (DTN_EPOCH - CCSDS_EPOCH).total_seconds()


def ccsdstime():
    """Obtains the current CCSDS timestamp.

    Returns:
        float: CCSDS timestamp
    """
    return time() + CCSDS_TO_UNIX_OFFSET


def sock_recv_raw(sock, count, timeout=None):
    """Tries to receive an exact number of bytes from a socket.

    Args:
        sock: socket object
        count (int): number of bytes to be received
        timeout (float): maximum time to wait, in s (None for infinite timeout)
    Returns:
        bytes: Received raw data
    """
    assert count
    buf = b""
    while len(buf) < count:
        if timeout:
            ready = select.select([sock], [], [], timeout)
        else:
            ready = select.select([sock], [], [])
        assert ready[0], "select operation ran into timeout"
        r = sock.recv(count - len(buf))
        assert len(r), "received 0 bytes (e.g. because the socket was closed)"
        buf += r
    return buf



"""BPbis bundle Generator

A simple Python package for generating and serialize BPbis bundles into CBOR.

Dependencies:
    This module depends on the ``cbor`` package which can be installed via pip:

    .. code:: bash

        pip install cbor
"""
import struct
import threading
from enum import IntEnum, IntFlag
from datetime import datetime, timezone, timedelta
from binascii import hexlify
import cbor  # type: ignore
from cbor.cbor import CBOR_ARRAY  # type: ignore
from .crc import crc16_x25, crc32_c
from .helpers import DTN_EPOCH, UNIX_TO_DTN_OFFSET


__all__ = [
    'BundleProcFlag',
    'CRCType',
    'BlockType',
    'BlockProcFlag',
    'RecordType',
    'ReasonCode',
    'StatusCode',
    'EID',
    'CreationTimestamp',
    'PrimaryBlock',
    'CanonicalBlock',
    'PayloadBlock',
    'AdministrativeRecord',
    'BundleStatusReport',
    'PreviousNodeBlock',
    'BundleAgeBlock',
    'HopCountBlock',
    'Bundle',
    'serialize_bundle7',
]


class BundleProcFlag(IntFlag):
    NONE = 0x000000
    IS_FRAGMENT = 0x000001
    ADMINISTRATIVE_RECORD = 0x000002
    MUST_NOT_BE_FRAGMENTED = 0x000004
    ACKNOWLEDGEMENT_REQUESTED = 0x000020
    REPORT_STATUS_TIME = 0x000040
    REPORT_RECEPTION = 0x004000
    REPORT_FORWARDING = 0x010000
    REPORT_DELIVERY = 0x020000
    REPORT_DELETION = 0x400000


class CRCType(IntEnum):
    NONE = 0   # indicates "no CRC is present."
    CRC16 = 1  # indicates "a CRC-16 (a.k.a., CRC-16-ANSI) is present."
    CRC32 = 2  # indicates "a standard IEEE 802.3 CRC-32 is present."


class BlockType(IntEnum):
    PRIMARY = -1
    PAYLOAD = 1
    PREVIOUS_NODE = 6
    BUNDLE_AGE = 7
    HOP_COUNT = 10


class BlockProcFlag(IntFlag):
    NONE = 0x00
    MUST_BE_REPLICATED = 0x01
    DISCARD_IF_UNPROC = 0x02
    REPORT_IF_UNPROC = 0x04
    DELETE_BUNDLE_IF_UNPROC = 0x10


class RecordType(IntEnum):
    BUNDLE_STATUS_REPORT = 1


class ReasonCode(IntEnum):
    """Bundle status report reason codes"""
    NO_INFO = 0
    LIFETIME_EXPIRDE = 1
    FORWARDED_UNIDIRECTIONAL = 2
    TRANSMISSION_CANCELED = 3
    DEPLETED_STORAGE = 4
    DEST_EID_UNINTELLIGIBLE = 5
    NO_KNOWN_ROUTE = 6
    NO_TIMELY_CONTACT = 7
    BLOCK_UNINTELLIGIBLE = 8
    HOP_LIMIT_EXCEEDED = 9
    TRAFFIC_PARED = 10


class StatusCode(IntEnum):
    RECEIVED_BUNDLE = 0x01
    FORWARDED_BUNDLE = 0x04
    DELIVERED_BUNDLE = 0x08
    DELETED_BUNDLE = 0x10


class EID(tuple):
    """BPbis Endpoint Identifier"""
    def __new__(cls, eid):
        # Tuples are immutable. Hence the construction logic is in the __new__
        # method instead in __init__.
        if eid is None or eid == 'dtn:none':
            return super().__new__(cls, (1, 0))

        # Copy existing EID
        if isinstance(eid, EID) or isinstance(eid, tuple):
            return super().__new__(cls, (eid[0], eid[1]))

        try:
            schema, ssp = eid.split(":")

            if schema == 'dtn':
                return super().__new__(cls, (1, ssp))
            elif schema == 'ipn':
                nodenum, servicenum = ssp.split('.')
                return super().__new__(
                    cls, (2, (int(nodenum), int(servicenum)))
                )
        except ValueError:
            raise ValueError("Invalid EID {!r}".format(eid))

        raise ValueError("Unknown schema {!r}".format(schema))

    @property
    def schema(self):
        """Schema (``dtn`` or ``ipn``) of the EID. This is encoded in the first
        tuple element."""
        if self[0] == 1:
            return 'dtn'
        elif self[0] == 2:
            return 'ipn'
        else:
            raise ValueError("Unknown schema {!r}".format(self[0]))

    @property
    def ssp(self):
        """Scheme-specific part (SSP) of the EID. This is encoded in the second
        tuple element.
        """
        if self[0] == 2:
            return "{}.{}".format(*self[1])
        else:
            if self[1] == 0:
                return "none"
            else:
                return self[1]

    def __str__(self):
        return "{}:{}".format(self.schema, self.ssp)

    def __repr__(self):
        # Reuse the __str__ method here
        return "<EID '{}'>".format(self)

    @staticmethod
    def from_cbor(cbor_data):
        assert len(cbor_data) == 2
        return EID((cbor_data[0], cbor_data[1]))


class CreationTimestamp(tuple):
    """BPbis creation timestamp tuple consisting of

        (<DTN timestamp>, <sequence number>)

    Args:
        time (None, int, datetime.datetime): Timestamp given as Unix timestamp
            (integer) or a Python timezone-aware datetime object. If the time
            is None, the current time will be used. If the time is 0, no
            conversion is performed and the DTN timestamp is set to 0.
        sequence_number (int): Sequence number of the bundle if the device is
            lacking a precise clock
    """
    def __new__(cls, time, sequence_number):
        if time == 0:
            return super().__new__(cls, [0, int(sequence_number)])

        # Use current datetime
        if time is None:
            time = datetime.now(timezone.utc)
        # Convert Unix timestamp into UTC datetime object
        elif isinstance(time, int) or isinstance(time, float):
            time = datetime.fromtimestamp(time, timezone.utc)

        return super().__new__(cls, [
            int(round((time - DTN_EPOCH).total_seconds() * 1000)),
            int(sequence_number),
        ])

    @property
    def time(self):
        """Returns the DTN timestamp value of the CreationTimestamp tuple as
        datetime object or None if it's 0.
        """
        return (
            None if self[0] == 0
            else DTN_EPOCH + timedelta(milliseconds=self[0])
        )

    @property
    def sequence_number(self):
        return self[1]

    def __repr__(self):
        return "<CreationTimestamp time={} sequence={}>".format(
                    self.time, self.sequence_number)

    @staticmethod
    def from_cbor(cbor_data):
        assert len(cbor_data) == 2
        return CreationTimestamp(
            cbor_data[0] / 1000 + UNIX_TO_DTN_OFFSET, cbor_data[1]
        )


class PrimaryBlock(object):
    """The primary bundle block contains the basic information needed to
    forward bundles to their destination.
    """
    def __init__(self, bundle_proc_flags=BundleProcFlag.NONE,
                 crc_type=CRCType.CRC16,
                 destination=None,
                 source=None,
                 report_to=None,
                 creation_time=None,
                 lifetime=24 * 60 * 60 * 1000,
                 fragment_offset=None,
                 total_payload_length=None,
                 crc_provided=None):
        self.version = 7
        self.bundle_proc_flags = bundle_proc_flags
        self.crc_type = crc_type
        self.destination = EID(destination)
        self.source = EID(source)
        self.report_to = EID(report_to)

        if not creation_time:
            creation_time = CreationTimestamp(datetime.now(timezone.utc), 0)

        self.creation_time = creation_time
        self.lifetime = lifetime

        # Optional fields
        self.fragment_offset = fragment_offset
        self.total_payload_length = total_payload_length

        self.crc_provided = crc_provided

    def has_flag(self, required):
        return self.bundle_proc_flags & required == required

    @property
    def block_number(self):
        return 0

    @property
    def block_type(self):
        return BlockType.PRIMARY

    def as_array(self):
        primary_block = [
            self.version,
            self.bundle_proc_flags,
            self.crc_type,
            self.destination,
            self.source,
            self.report_to,
            self.creation_time,
            self.lifetime,
        ]

        # Fragmentation
        if self.has_flag(BundleProcFlag.IS_FRAGMENT):

            assert self.fragment_offset is not None, (
                "Fragment offset must be present for fragmented bundles"
            )
            assert self.total_payload_length is not None, (
                "Total payload length must be present for fragmented bundles"
            )

            primary_block.append(self.fragment_offset)
            primary_block.append(self.total_payload_length)

        return primary_block

    def calculate_crc(self):
        primary_block = self.as_array()
        assert self.crc_type != CRCType.NONE
        binary = [
            # CBOR Array header
            struct.pack('B', 0x80 | len(primary_block) + 1),
        ]

        # encode each field
        for field in primary_block:
            binary.append(cbor.dumps(field))

        if self.crc_type == CRCType.CRC32:
            # Empty CRC-32 bit field: CBOR "byte string"
            binary.append(b'\x44\x00\x00\x00\x00')
            crc = crc32_c(b''.join(binary))
        else:
            binary.append(b'\x42\x00\x00')
            crc = crc16_x25(b''.join(binary))

        return crc

    def __bytes__(self):
        primary_block = self.as_array()

        if self.crc_type != CRCType.NONE:
            crc = self.calculate_crc()
            if self.crc_type == CRCType.CRC32:
                primary_block.append(struct.pack('!I', crc))
            else:
                primary_block.append(struct.pack('!H', crc))

        return cbor.dumps(primary_block)

    def __repr__(self):
        return (
            "PrimaryBlock(bundle_proc_flags={}, crc_type={}, "
            "destination={}, source={}, report_to={}, "
            "creation_time={}, lifetime={}, fragment_offset={}, "
            "total_payload_length={}, crc_provided={})"
        ).format(
            repr(self.bundle_proc_flags),
            repr(self.crc_type),
            repr(self.destination),
            repr(self.source),
            repr(self.report_to),
            self.creation_time,
            self.lifetime,
            self.fragment_offset,
            self.total_payload_length,
            (
                "0x{:08x}".format(self.crc_provided)
                if self.crc_provided is not None else "None"
            ),
        )

    @staticmethod
    def from_cbor(cbor_data):
        version = cbor_data[0]
        assert version == 7, "This class can only decode BPv7 bundles."
        bundle_proc_flags = BundleProcFlag(cbor_data[1])
        crc_type = CRCType(cbor_data[2])

        expected_fields = 8

        fragment_offset = None
        total_payload_length = None
        if (bundle_proc_flags & BundleProcFlag.IS_FRAGMENT) != 0:
            expected_fields += 2
            fragment_offset = cbor_data[8]
            total_payload_length = cbor_data[9]

        crc = None
        if crc_type != CRCType.NONE:
            crc_data = cbor_data[expected_fields]  # obtain last field of array
            expected_fields += 1
            if crc_type == CRCType.CRC32:
                crc = struct.unpack('!I', crc_data)[0]
            else:
                crc = struct.unpack('!H', crc_data)[0]

        return PrimaryBlock(
            bundle_proc_flags=bundle_proc_flags,
            crc_type=crc_type,
            destination=EID.from_cbor(cbor_data[3]),
            source=EID.from_cbor(cbor_data[4]),
            report_to=EID.from_cbor(cbor_data[5]),
            creation_time=CreationTimestamp.from_cbor(cbor_data[6]),
            fragment_offset=fragment_offset,
            total_payload_length=total_payload_length,
            lifetime=(cbor_data[7]),
            crc_provided=crc,
        )


class CanonicalBlock(object):
    """Canonical bundle block structure"""
    def __init__(self, block_type, data,
                 block_number=None,
                 block_proc_flags=BlockProcFlag.NONE,
                 crc_type=CRCType.CRC32,
                 crc_provided=None):
        self.block_type = block_type
        self.block_proc_flags = block_proc_flags
        self.block_number = block_number
        self.crc_type = crc_type
        self.crc_provided = crc_provided
        self.data = data

    def as_array(self):
        return [
            self.block_type,
            self.block_number,
            self.block_proc_flags,
            self.crc_type,
            self.data,
        ]

    def calculate_crc(self):
        assert self.crc_type != CRCType.NONE
        block = self.as_array()
        if self.crc_type == CRCType.CRC32:
            # Empty CRC-32 bit field: CBOR "byte string"
            empty_crc = b'\x44\x00\x00\x00\x00'
        else:
            empty_crc = b'\x42\x00\x00'

        binary = struct.pack('B', 0x80 | (len(block) + 1)) + b''.join(
            [cbor.dumps(item) for item in block]
        ) + empty_crc

        if self.crc_type == CRCType.CRC32:
            crc = crc32_c(binary)
            # unsigned long in network byte order
            block.append(struct.pack('!I', crc))
        else:
            crc = crc16_x25(binary)
            # unsigned short in network byte order
            block.append(struct.pack('!H', crc))
        return crc

    def __bytes__(self):
        assert isinstance(self.data, bytes)
        block = self.as_array()

        if self.crc_type != CRCType.NONE:
            crc = self.calculate_crc()
            if self.crc_type == CRCType.CRC32:
                block.append(struct.pack('!I', crc))
            else:
                block.append(struct.pack('!H', crc))

        return cbor.dumps(block)

    def __repr__(self):
        return (
            "{}(block_type={}, block_number={}, block_proc_flags={}, "
            "data_length={}, crc_type={}, crc_provided={})"
        ).format(
            self.__class__.__name__,
            repr(self.block_type),
            repr(self.block_number),
            repr(self.block_proc_flags),
            len(self.data),
            repr(self.crc_type),
            (
                "0x{:08x}".format(self.crc_provided)
                if self.crc_provided is not None else "None"
            ),
        )

    @staticmethod
    def from_cbor(cbor_data):
        crc = None
        crc_type = CRCType(cbor_data[3])
        if crc_type == CRCType.NONE:
            assert len(cbor_data) == 5
        elif crc_type == CRCType.CRC32:
            assert len(cbor_data) == 6
            crc = struct.unpack('!I', cbor_data[5])[0]
        elif crc_type == CRCType.CRC16:
            assert len(cbor_data) == 6
            crc = struct.unpack('!H', cbor_data[5])[0]
        else:
            assert False

        return CanonicalBlock(
            block_type=BlockType(cbor_data[0]),
            data=cbor_data[4],
            block_number=cbor_data[1],
            block_proc_flags=BlockProcFlag(cbor_data[2]),
            crc_type=crc_type,
            crc_provided=crc,
        )


class PayloadBlock(CanonicalBlock):

    def __init__(self, data, **kwargs):
        super().__init__(BlockType.PAYLOAD,
                         data,
                         block_number=1,
                         **kwargs)


class CBORBlock(CanonicalBlock):

    def __init__(self, block_type, cbor_data, **kwargs):
        super().__init__(block_type, cbor.dumps(cbor_data), **kwargs)


# ----------------------
# Administrative Records
# ----------------------

class AdministrativeRecord(PayloadBlock):

    def __init__(self, bundle, record_type_code, record_data):
        record_data.extend([
            bundle.primary_block.source,
            bundle.primary_block.creation_time
        ])

        if bundle.is_fragmented:
            record_data.extend([
                bundle.primary_block.fragment_offset,
                bundle.primary_block.total_payload_length
            ])

        super().__init__(data=cbor.dumps([
            record_type_code,
            record_data
        ]))


class BundleStatusReport(AdministrativeRecord):

    def __init__(self, infos, reason, bundle, time=None):
        status_info = [
            [infos & StatusCode.RECEIVED_BUNDLE != 0],
            [infos & StatusCode.FORWARDED_BUNDLE != 0],
            [infos & StatusCode.DELIVERED_BUNDLE != 0],
            [infos & StatusCode.DELETED_BUNDLE != 0],
        ]

        if bundle.primary_block.has_flag(BundleProcFlag.REPORT_STATUS_TIME):
            if not time:
                time = datetime.now(timezone.utc)
            for info in status_info:
                if info[0]:
                    info.append(int(round((time - DTN_EPOCH).total_seconds())))

        record_data = [
            status_info,
            reason,
        ]

        super().__init__(
            bundle,
            record_type_code=RecordType.BUNDLE_STATUS_REPORT,
            record_data=record_data
        )

    @property
    def status_info(self):
        return self.data[1][0]

    def __repr__(self):
        return "<BundleStatusReport {!r}>".format(self.status_info)


# ----------------
# Extension Blocks
# ----------------

class PreviousNodeBlock(CBORBlock):

    def __init__(self, eid, **kwargs):
        super().__init__(BlockType.PREVIOUS_NODE, EID(eid), **kwargs)


class BundleAgeBlock(CBORBlock):
    """The Bundle Age block, block type 7, contains the number of milliseconds
    that have elapsed between the time the bundle was created and time at which
    it was most recently forwarded.

    Args:
        age (int): Age value in seconds
    """

    def __init__(self, age, **kwargs):
        super().__init__(BlockType.BUNDLE_AGE, int(age * 1000), **kwargs)


class HopCountBlock(CBORBlock):

    def __init__(self, hop_limit, hop_count, **kwargs):
        super().__init__(BlockType.HOP_COUNT, (hop_limit, hop_count), **kwargs)


class BundleMeta(type):
    """Metaclass for :class:`Bundle` providing the special initialization
    capability of a bundle object with a block list.

    This metaclass deconstructs the given block list to match the `__init__()`
    method of :class:`Bundle`. The first list element is treated as primary
    block, the last element as payload block. Every block in between these
    blocks are handled as extension block.
    """
    def __call__(cls, *args, **kwargs):
        # Special case:
        #     Bundle is created from a single list of blocks.
        #     Interpret the list elements by their position.
        if len(args) == 1:
            # Ensure there are enough block elements
            if len(args[0]) < 2:
                raise TypeError(
                    "Block list must contain at least two elements"
                )

            primary_block = args[0][0]     # First element
            payload_block = args[0][-1]    # Last element
            blocks = args[0][1:-1]         # Everything in between

            # Rearrange arguments to fit constructor
            args = (primary_block, payload_block, blocks)

        return super().__call__(*args, **kwargs)


class Bundle(object, metaclass=BundleMeta):
    """BPbis bundle data structure used for serialization

    A bundle is a list of blocks. The first list element is the primary block,
    the last list element is the payload block. Extension blocks are in between
    these two blocks.

    Args:
        primary_block (PrimaryBlock): Headers of the bundle
        payload_block (PayloadBlock): Payload of the bundle
        blocks (List[CanonicalBlock], optional): List of optional canonical
            (extension) blocks.

    There is a special constructor where you can create a bundle from a list of
    blocks. The blocks list is interpreted by there position in the list as
    desribed above. This capability is provided by the :class:`BundleMeta` meta
    class.

    .. code:: python

        bundle = Bundle([
            primary_block,
            hop_count,
            payload_block
        ])

    The :class:`Bundle` class supports the iterator protocol to iterate over
    every block in the same order as it would be encoded in CBOR.

    ..code:: python

        for i, block in enumerate(bundle):
            if i == 0:
                assert isinstance(block, PrimaryBlock)
    """
    def __init__(self, primary_block, payload_block, blocks=None):
        assert (
            primary_block.creation_time.time or
            blocks and
            any(b.block_type == BlockType.BUNDLE_AGE for b in blocks)
        ), "There must be a 'Bundle Age' block if the creation time is 0"

        self.primary_block = primary_block
        self.payload_block = payload_block
        self.blocks = []

        if blocks:
            for block in blocks:
                self.add(block)

    @property
    def is_fragmented(self):
        return self.primary_block.has_flag(BundleProcFlag.IS_FRAGMENT)

    def add(self, new_block):
        num = 1

        for block in self.blocks:
            message = "Block {} number already assigned".format(
                block.block_number
            )

            assert block.block_number != new_block.block_number, message

            # Search for a new block number if the block does not already
            # contain a block number
            num = max(num, block.block_number)

            # Assert unique block types
            if block.block_type == new_block.block_type:
                # Previous Node
                if block.block_type == BlockType.PREVIOUS_NODE:
                    raise ValueError(
                        "There must be only one 'Previous Node' block"
                    )
                # Hop Count
                elif block.block_type == BlockType.HOP_COUNT:
                    raise ValueError(
                        "There must be only one 'Hop Count' block"
                    )
                # Bundle Age
                elif block.block_type == BlockType.BUNDLE_AGE:
                    raise ValueError(
                        "There must be only one 'Bundle Age' block"
                    )

        if new_block.block_number is None:
            new_block.block_number = num + 1

        # Previous Node blocks must be the first blocks after the primary block
        if new_block.block_type == BlockType.PREVIOUS_NODE:
            self.blocks.insert(0, new_block)
        else:
            self.blocks.append(new_block)

    def hexlify(self):
        """Return the hexadecimal representation of the CBOR encoded bundle.

        Returns:
            bytes: CBOR encoded bundle as hex string
        """
        return hexlify(bytes(self))

    def __iter__(self):
        yield self.primary_block
        yield from self.blocks
        yield self.payload_block

    def __bytes__(self):
        # Header for indefinite array
        head = struct.pack('B', CBOR_ARRAY | 31)
        # Stop-code for indefinite array
        stop = b'\xff'
        return head + b''.join(bytes(block) for block in self) + stop

    def __repr__(self):
        return "{}(primary_block={}, payload_block={}, blocks=[{}])".format(
            self.__class__.__name__,
            repr(self.primary_block),
            repr(self.payload_block),
            ", ".join([repr(b) for b in self.blocks]),
        )

    @staticmethod
    def parse(data):
        cbor_data = cbor.loads(data)
        # At least a primary and a payload block are required
        assert len(cbor_data) >= 2
        primary_block = PrimaryBlock.from_cbor(cbor_data[0])
        payload_block = PayloadBlock.from_cbor(cbor_data[-1])
        blocks = [CanonicalBlock.from_cbor(e) for e in cbor_data[1:-1]]
        return Bundle(primary_block, payload_block, blocks)


_th_local = threading.local()


def next_sequence_number():
    seqnum = _th_local.__dict__.get("seqnum", 0)
    _th_local.__dict__["seqnum"] = seqnum + 1
    return seqnum


def reset_sequence_number():
    _th_local.__dict__["seqnum"] = 0


def create_bundle7(source_eid, destination_eid, payload,
                   report_to_eid=None, crc_type_primary=CRCType.CRC32,
                   creation_timestamp=None, sequence_number=None,
                   lifetime=300, flags=BlockProcFlag.NONE,
                   fragment_offset=None, total_adu_length=None,
                   hop_limit=None, hop_count=0, bundle_age=None,
                   previous_node_eid=None,
                   crc_type_canonical=CRCType.CRC16):
    """All-in-one function to encode a payload from a source EID
    to a destination EID as BPbis bundle.

    Args:
        source_eid (EID, str): Source endpoint address
        destination_eid (EID, str): Destination endpoint address
        report_to_eid (EID, str, optional): Endpoint address that should
            receive bundle status reports. If not given the null EID will be
            used.
        crc_type_primary (CRCType, optional): The kind of CRC used for the
            primary block.
        creation_timestamp (datetime, int, optional): Unix timestamp or
            timezone-aware datetime object when the bundle was created. If not
            given the current timestamp will be used.
        sequence_number (int, optional): Sequence number that is used for the
            bundle. If the device lacks a precise clock, this is the only
            source of information for differentiating two subsequent bundles.
            If None, the last thread-local sequence number will be used,
            incremented by one.
        lifetime (int, optional): Bundle lifetime in seconds
        flags (BlockProcFlag, optional): Bundle processing flags
        fragment_offset (int, optional): If the bundle is fragmented, use this
            offset. This value is only used if the bundle processing flag
            :attr:`BundleProcFlag.IS_FRAGMENT` is set.
        total_adu_length (int, optional): If the bundle is fragmented, use this
            to specify the total data length. This value is only used if the
            bundle processing flag :attr:`BundleProcFlag.IS_FRAGMENT` is set.
        hop_limit (int, optional): Maximal number of hops (intermediate DTN
            nodes) the bundle is allowed to use to reach its destination.
        hop_count (int, optional): Current hop count
        bundle_age (int, optional): Age of the bundle in seconds
        previous_node_eid (EID, str, optional): Address of the previous
            endpoint the bundle was received from Returns: bytes: CBOR encoded
            BPbis bundle
        crc_type_canonical (CRCType, optional): The kind of CRC used for the
            canonical blocks.
    """
    blocks = []

    if hop_limit is not None and hop_count is not None:
        blocks.append(
            HopCountBlock(hop_limit, hop_count, crc_type=crc_type_canonical)
        )
    if previous_node_eid is not None:
        blocks.append(
            PreviousNodeBlock(previous_node_eid, crc_type=crc_type_canonical)
        )
    if bundle_age is not None:
        blocks.append(
            BundleAgeBlock(bundle_age, crc_type=crc_type_canonical)
        )

    return Bundle(
        PrimaryBlock(
            bundle_proc_flags=flags,
            crc_type=crc_type_primary,
            destination=destination_eid,
            source=source_eid,
            report_to=report_to_eid,
            creation_time=CreationTimestamp(
                creation_timestamp,
                (
                    sequence_number
                    if sequence_number is not None
                    else next_sequence_number()
                ),
            ),
            lifetime=lifetime * 1000,
            fragment_offset=fragment_offset,
            total_payload_length=total_adu_length,
        ),
        PayloadBlock(
            payload,
            crc_type=crc_type_canonical,
        ),
        blocks,
    )


def serialize_bundle7(source_eid, destination_eid, payload,
                      report_to_eid=None, crc_type_primary=CRCType.CRC32,
                      creation_timestamp=None, sequence_number=None,
                      lifetime=300, flags=BlockProcFlag.NONE,
                      fragment_offset=None, total_adu_length=None,
                      hop_limit=None, hop_count=0, bundle_age=None,
                      previous_node_eid=None,
                      crc_type_canonical=CRCType.CRC16):
    """All-in-one function to encode a payload from a source EID
    to a destination EID as BPbis bundle.
    See create_bundle7 for a description of options."""
    return bytes(create_bundle7(
        source_eid, destination_eid, payload,
        report_to_eid, crc_type_primary,
        creation_timestamp, sequence_number,
        lifetime, flags,
        fragment_offset, total_adu_length,
        hop_limit, hop_count, bundle_age,
        previous_node_eid,
        crc_type_canonical
    ))



import enum
import struct
import socket
import logging
from datetime import datetime, timedelta, timezone

from .crc import crc16_ccitt_false
from .helpers import CCSDS_EPOCH, sock_recv_raw

logger = logging.getLogger(__name__)

SPP_MAX_APID = 0x7ff
SPP_MAX_SEGMENT_NUMBER = 0x3fff
SPP_MAX_DATA_LENGTH = 65536

SPP_PRIMARY_HEADER_LENGTH = 6
SPP_PH_P1_VERSION_MASK = 0xe000
SPP_PH_EARLY_VERSION_MASK = SPP_PH_P1_VERSION_MASK >> 8
SPP_PH_P1_TYPE_MASK = 0x1000
SPP_PH_P1_SECONDARY_HEADER_FLAG_MASK = 0x0800
SPP_PH_P1_APID_MASK = 0x07ff
SPP_PH_P2_SEQUENCE_FLAGS_MASK = 0xc000
SPP_PH_P2_SEQUENCE_FLAGS_SHIFT = 14
SPP_PH_P2_SEQUENCE_CNT_MASK = 0x3fff

SPP_TC_UNSEG_BASE_UNIT_LONGP_THRESHOLD = 4
SPP_TC_UNSEG_FRAC_LONGP_THRESHOLD = 3


class SPPSegmentStatus(enum.IntFlag):
    ContinuationSegment = 0,
    FirstSegment = 1,
    LastSegment = 2,
    Unsegmented = 3,


class SPPPrimaryHeaderFlags(enum.IntFlag):
    HasSecondaryHeader = SPP_PH_P1_SECONDARY_HEADER_FLAG_MASK,
    PacketTypeIsRequest = SPP_PH_P1_TYPE_MASK,


class SPPTimecodeType(enum.IntEnum):
    UnsegmentedCCSDSEpoch = 0x1,
    UnsegmentedCustomEpoch = 0x2,
    DaySegmented = 0x4,
    CalendarSegmented = 0x5,
    Custom = 0x6,
    Unknown = 0x7,


class SPPTimecode(object):

    def __init__(self, dt=None, fractional=0,
                 timecode_type=SPPTimecodeType.UnsegmentedCCSDSEpoch,
                 base_unit_octets=4, fractional_octets=0,
                 has_preamble=True):
        self.dt = dt or datetime.now(timezone.utc)
        self.fractional = fractional
        self.timecode_type = timecode_type
        self.base_unit_octets = base_unit_octets
        self.fractional_octets = fractional_octets
        self.has_preamble = has_preamble

    def length(self):
        return (
            (1 if self.has_preamble else 0) +
            self.base_unit_octets +
            self.fractional_octets
        )

    def __bytes__(self):
        assert self.has_preamble

        # We do not support other configurations for now.
        assert self.timecode_type == SPPTimecodeType.UnsegmentedCCSDSEpoch

        # We do not allow timecodes longer than what can be represented in
        # the preamble.
        assert self.base_unit_octets > 0
        assert self.fractional_octets >= 0
        assert self.base_unit_octets <= SPP_TC_UNSEG_BASE_UNIT_LONGP_THRESHOLD
        assert self.fractional_octets <= SPP_TC_UNSEG_FRAC_LONGP_THRESHOLD
        preamble = 0b0_000_00_00  # second preamble bit = 0b0
        preamble |= int(self.timecode_type) << 4
        preamble |= (self.base_unit_octets - 1) << 2
        preamble |= self.fractional_octets

        ret = bytes([preamble])
        ret += int((self.dt - CCSDS_EPOCH).total_seconds()).to_bytes(
            self.base_unit_octets,
            "big"
        )
        if self.fractional_octets:
            ret += int(self.fractional).to_bytes(
                self.fractional_octets,
                "big"
            )

        return ret

    @staticmethod
    def parse(data, has_preamble=True):
        assert has_preamble

        preamble = data[0]
        timecode_type = (preamble & 0b0_111_00_00) >> 4
        base_unit_octets = ((preamble & 0b0_000_11_00) >> 2) + 1
        fractional_octets = (preamble & 0b0_000_00_11)
        assert timecode_type == SPPTimecodeType.UnsegmentedCCSDSEpoch

        fractional_offset = 1 + base_unit_octets
        fractional_end = fractional_offset + fractional_octets

        dt = CCSDS_EPOCH + timedelta(seconds=int.from_bytes(
            data[1:fractional_offset],
            "big"
        ))
        fractional = int.from_bytes(
            data[fractional_offset:fractional_end],
            "big"
        )

        return SPPTimecode(
            dt=dt,
            fractional=fractional,
            timecode_type=SPPTimecodeType(timecode_type),
            base_unit_octets=base_unit_octets,
            fractional_octets=fractional_octets,
        ), fractional_end

    def __repr__(self):
        return (
            "SPPTimecode(dt={}, fractional={}, timecode_type={}, "
            "base_unit_octets={}, fractional_octets={}, has_preamble={})"
        ).format(
            repr(self.dt),
            self.fractional,
            repr(self.timecode_type),
            self.base_unit_octets,
            self.fractional_octets,
            self.has_preamble,
        )


class SPPPacketHeader(object):

    def __init__(self, payload_length=0, is_request=False, apid=0x1,
                 segment_status=SPPSegmentStatus.Unsegmented,
                 segment_number=0, timecode=None, ancillary_data=None):
        self.payload_length = payload_length
        self.is_request = is_request
        assert apid >= 0
        assert apid <= SPP_MAX_APID
        self.apid = apid
        self.segment_status = segment_status
        assert segment_number >= 0
        assert segment_number <= SPP_MAX_SEGMENT_NUMBER
        self.segment_number = segment_number
        self.timecode = timecode
        self.ancillary_data = ancillary_data

    def data_length(self):
        return self.payload_length + self.secondary_header_length()

    def secondary_header_length(self):
        return (
            len(self.ancillary_data) if self.ancillary_data else 0 +
            self.timecode.length() if self.timecode else 0
        )

    def total_header_length(self):
        return SPP_PRIMARY_HEADER_LENGTH + self.secondary_header_length()

    def minimum_payload_length(self):
        if self.secondary_header_length() > 0:
            return 0
        return 1

    def __bytes__(self):
        assert self.payload_length >= self.minimum_payload_length()

        data_length = self.data_length()
        assert data_length <= SPP_MAX_DATA_LENGTH

        part1 = self.apid
        if self.is_request:
            part1 |= SPP_PH_P1_TYPE_MASK
        # Has secondary header?
        if self.ancillary_data is not None or self.timecode is not None:
            part1 |= SPP_PH_P1_SECONDARY_HEADER_FLAG_MASK
        part2 = (
            (self.segment_status << SPP_PH_P2_SEQUENCE_FLAGS_SHIFT) |
            self.segment_number
        )

        primary_header = [
            struct.pack('!H', part1),
            struct.pack('!H', part2),
            struct.pack('!H', self.data_length() - 1),
        ]

        secondary_header = []

        if self.timecode is not None:
            secondary_header.append(bytes(self.timecode))

        # NOTE that the length of the ancillary data field has to be fixed!
        if self.ancillary_data is not None:
            secondary_header.append(bytes(self.ancillary_data))

        return b"".join(primary_header + secondary_header)

    @staticmethod
    def preparse_data_length(primary_header_bytes):
        part1, _, data_length = struct.unpack('!HHH', primary_header_bytes[:6])
        has_secondary_header = (
            part1 & SPP_PH_P1_SECONDARY_HEADER_FLAG_MASK
        ) != 0
        return data_length + 1, has_secondary_header

    @staticmethod
    def parse(data, timecode_used=False, ancillary_data_length=0):
        part1, part2, data_length = struct.unpack('!HHH', data[:6])

        assert (part1 & 0b1110_0000_0000_0000) == 0
        is_request = (part1 & SPP_PH_P1_TYPE_MASK) != 0
        has_secondary_header = (
            part1 & SPP_PH_P1_SECONDARY_HEADER_FLAG_MASK
        ) != 0
        apid = part1 & SPP_PH_P1_APID_MASK

        segment_status = (
            part2 & SPP_PH_P2_SEQUENCE_FLAGS_MASK
        ) >> SPP_PH_P2_SEQUENCE_FLAGS_SHIFT
        segment_number = part2 & SPP_PH_P2_SEQUENCE_CNT_MASK

        data_length += 1
        payload_offset = SPP_PRIMARY_HEADER_LENGTH

        timecode = None
        ancillary_data = None
        if has_secondary_header:
            if timecode_used:
                assert has_secondary_header
                timecode, timecode_len = SPPTimecode.parse(
                    data[payload_offset:]
                )
                payload_offset += timecode_len
            if ancillary_data_length:
                assert has_secondary_header
                ancillary_data_start = payload_offset
                payload_offset += ancillary_data_length
                ancillary_data = data[ancillary_data_start:payload_offset]

        payload_length = (
            data_length - (payload_offset - SPP_PRIMARY_HEADER_LENGTH)
        )

        return SPPPacketHeader(
            payload_length=payload_length,
            is_request=is_request,
            apid=apid,
            segment_status=SPPSegmentStatus(segment_status),
            segment_number=segment_number,
            timecode=timecode,
            ancillary_data=ancillary_data,
        ), payload_offset

    def __repr__(self):
        return (
            "SPPPacketHeader(payload_length={}, is_request={}, apid=0x{:02x}, "
            "segment_status={}, segment_number={}, timecode={}, "
            "ancillary_data={})"
        ).format(
            self.payload_length,
            self.is_request,
            self.apid,
            self.segment_status,
            self.segment_number,
            self.timecode,
            "<len={}>".format(
                len(self.ancillary_data)
            ) if self.ancillary_data else "None",
        )


class SPPPacket(object):

    def __init__(self, header, payload, has_crc=False, crc_provided=None):
        header.payload_length = len(payload)
        if has_crc:
            header.payload_length += 2
        self.header = header
        self.payload = payload
        self.has_crc = has_crc
        self.crc_provided = crc_provided

    def crc(self, binary=None):
        return crc16_ccitt_false(binary or (bytes(self.header) + self.payload))

    def __bytes__(self):
        packet = bytes(self.header) + self.payload
        if self.has_crc:
            packet += struct.pack("!H", self.crc(packet))
        return packet

    @staticmethod
    def parse(data, has_crc=False, timecode_used=False,
              ancillary_data_length=0):
        header, header_end = SPPPacketHeader.parse(
            data,
            timecode_used=timecode_used,
            ancillary_data_length=ancillary_data_length,
        )

        payload_end = header_end + header.payload_length
        payload = data[header_end:payload_end]

        crc = None
        if has_crc:
            crc = struct.unpack("!H", payload[-2:])[0]
            payload = payload[:-2]

        return SPPPacket(header, payload, has_crc, crc), payload_end

    def __repr__(self):
        return (
            "SPPPacket(header={}, payload=<length={}>, has_crc={}, "
            "crc_provided={})"
        ).format(
            repr(self.header),
            len(self.payload),
            self.has_crc,
            "None" if self.crc_provided is None else "<0x{:04x}, {}>".format(
                self.crc_provided,
                "OK" if self.crc() == self.crc_provided else "FAIL",
            ),
        )


class TCPSPPConnection(object):
    """TCPSPP connection to a remote DTN node.

    The connection supports the context manager protocol to ensure a proper
    shutdown of the connection in case of any error:

    .. code:: python

        from pyd3tn.spp import TCPSPPConnection

        with TCPSPPConnection('127.0.0.1', 4223) as conn:
            conn.send(bundle)

    Args:
        host (str): Host to connect to
        port (int): TCP port to connect to
        use_crc (bool): Whether or not the SPP CRC is used
        timeout (float): Timeout for receiving bundles (None for infinite)
    """
    def __init__(self, host, port, use_crc=False, timeout=None):
        self.host = host
        self.port = port
        self.use_crc = use_crc
        self.timeout = timeout
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

    def __enter__(self):
        self.connect()
        return self

    def __exit__(self, *args):
        self.disconnect()

    def connect(self):
        """Establish a TCPSPP connection to a remote peer."""
        self.sock.connect((self.host, self.port))
        self.sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
        logger.debug("TCPSPP: Connected to %s:%d", self.host, self.port)

    def disconnect(self):
        """Close a TCPSPP connection."""
        logger.debug(
            "TCPSPP: Terminate connection to %s:%d",
            *self.sock.getpeername(),
        )
        self.sock.shutdown(socket.SHUT_RDWR)
        self.sock.close()

    def send_bundle(self, bundle):
        """Send a single bundle to the remote peer.

        Args:
            bundle (bytes): Serialized bundle
        """
        raw = bytes(SPPPacket(
            SPPPacketHeader(
                timecode=SPPTimecode(),
            ),
            bundle,
            has_crc=self.use_crc,
        ))
        return self.sock.sendall(raw)

    def recv_bundle(self):
        """Receives a single bundle from the remote peer.

        Returns:
            bundle (bytes): Serialized bundle
        """
        hdr_raw = sock_recv_raw(self.sock, 6, self.timeout)
        data_length, has_secondary = SPPPacketHeader.preparse_data_length(
            hdr_raw
        )
        data_plus_secondary = sock_recv_raw(
            self.sock,
            data_length,
            self.timeout,
        )
        packet, _ = SPPPacket.parse(
            hdr_raw + data_plus_secondary,
            timecode_used=True,
            has_crc=self.use_crc,
        )
        if self.use_crc:
            assert packet.crc_provided == packet.crc()
        return packet.payload



import enum
import struct
import socket
import logging
import asyncio

from .sdnv import sdnv_encode, sdnv_decode
from .helpers import sock_recv_raw


logger = logging.getLogger(__name__)


class TCPCLConnectionFlag(enum.IntFlag):
    NONE = 0x00
    REQUEST_ACK = 0x01
    REACTIVE_FRAGMENTATION = 0x02
    ALLOW_REFUSAL = 0x04
    REQUEST_LENGTH = 0x08
    DEFAULT = NONE


def serialize_tcpcl_contact_header(source_eid,
                                   flags=TCPCLConnectionFlag.DEFAULT,
                                   keepalive_interval=0):
    # https://tools.ietf.org/html/rfc7242#section-4.1
    result = bytearray()
    result += b"dtn!"  # TCPCL header magic
    result += b"\x03"  # version = v3
    result.append(flags)
    result += struct.pack("!H", keepalive_interval)
    source_eid_bin = source_eid.encode("ascii")
    result += sdnv_encode(len(source_eid_bin))
    result += source_eid_bin
    return result


def decode_tcpcl_contact_header(header):
    assert header[0:4] == b"dtn!", "header magic not found"
    assert header[4] == 0x03, "only TCPCL v3 is supported"
    assert len(header) >= 9, "corrupt (too short) TCPCL header"
    result = {
        "version": header[4],
        "flags": header[5],
        "keepalive_interval": struct.unpack('!H', header[6:8])[0],
    }
    eid_len, end_offset = sdnv_decode(header[8:])
    assert len(header) == 8 + end_offset + eid_len, "wrong header length"
    result["eid"] = header[8 + end_offset:].decode("ascii")
    return result


def serialize_tcpcl_single_bundle_segment(bundle_data):
    # https://tools.ietf.org/html/rfc7242#section-5.2
    result = bytearray()
    message_type = 0x1  # DATA_SEGMENT
    message_flags = 0x3  # S | E = first AND last segment of bundle
    result.append((message_type << 4) | message_flags)
    result += sdnv_encode(len(bundle_data))
    result += bundle_data
    return result


def decode_tcpcl_bundle_segment_header(bindata):
    message_type = (bindata[0] & 0xF0) >> 4
    message_flags = bindata[0] & 0x0F
    assert message_type == 0x1, "only DATA_SEGMENT messages are supported"
    assert message_flags == 0x3, "only unsegmented bundles are supported"
    length, _ = sdnv_decode(bindata[1:])
    return message_type, message_flags, length


class TCPCLConnection(object):
    """TCPCL connection to a remote DTN node

    The connection supports the context manager protocol to ensure a proper
    shutdown of the connection in case of any error:

    .. code:: python

        from pyd3tn.tcpcl import TCPCLConnection

        with TCPCLConnection('dtn://my-eid.dtn', '127.0.0.1', 4223) as conn:
            conn.send(bundle)

    Args:
        eid (str): Endpoint identifier for this DTN node
        host (str): Host to connect to
        port (int): TCP port to connect to
        timeout (float): Timeout for receiving bundles (None for infinite)
    """
    def __init__(self, eid, host, port, timeout=None):
        self.eid = eid
        self.host = host
        self.port = port
        self.timeout = timeout
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

    def __enter__(self):
        self.connect()
        return self

    def __exit__(self, *args):
        self.disconnect()

    def connect(self):
        """Establish a TCPCL connection to a remote peer.

        Returns:
            dict: Deserialized TCPCL contact header
        """
        if not self.sock:
            return

        # Establish TCP connection
        self.sock.connect((self.host, self.port))
        self.sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
        logger.debug("TCPCL: Connected to %s:%d", self.host, self.port)

        # Send TCPCL contact header with own EID
        logger.debug("TCPCL: Send header for %r", self.eid)
        self.sock.sendall(serialize_tcpcl_contact_header(self.eid))

        # Receive TCPCL contact header
        header = decode_tcpcl_contact_header(self.sock.recv(1024))
        logger.debug("TCPCL: Received header: %r", header)

        return header

    def disconnect(self):
        # Terminate TCPCL connection
        logger.debug(
            "TCPCL: Terminate connection to %s:%d",
            *self.sock.getpeername()
        )
        self.sock.sendall(b"\x50")
        self.sock.shutdown(socket.SHUT_RDWR)
        self.sock.close()

    def send_bundle(self, bundle):
        """Sends a single bundle to the remote peer.

        Args:
            bundle (bytes): Serialized bundle
        """
        return self.sock.sendall(serialize_tcpcl_single_bundle_segment(bundle))

    def recv_bundle(self):
        """Receives a single bundle from the remote peer.

        Returns:
            bundle (bytes): Serialized bundle
        """
        hdr_raw = sock_recv_raw(self.sock, 2, self.timeout)
        c = 0
        # Incrementally read length until end of SDNV is detected (limit = 10)
        while (hdr_raw[-1] & 0x80) != 0 and c < 10:
            hdr_raw += sock_recv_raw(self.sock, 1, self.timeout)
            c += 1
        assert (hdr_raw[-1] & 0x80) == 0, (
            "invalid header or SDNV longer than 10 bytes"
        )
        type_, flags, length = decode_tcpcl_bundle_segment_header(hdr_raw)
        data = sock_recv_raw(
            self.sock,
            length,
            self.timeout,
        )
        return data


class AsyncTCPCLConnection(TCPCLConnection):
    """Asyncio-variant of the TCPCL connection

    It supports the async context manager protocol:

    .. code:: python

        async with AsyncTCPCLConnection(me) as conn:
            conn.bind('127.0.0.1', 42421)
            await conn.connect()
            await conn.send(bundle)

    Args:
        eid (str): Endpoint identifier for this DTN node
        loop (asyncio.AbstractEventLoop, optional): Event loop that should be
            used. If not given the default :func:`asyncio.get_event_loop` will
            be used.
    """

    def __init__(self, eid, loop=None):
        super().__init__(eid)

        self.loop = loop or asyncio.get_event_loop()
        self.reader = None
        self.writer = None

    async def __aenter__(self):
        self.__enter__()
        self.sock.setblocking(False)
        return self

    async def connect(self, host='127.0.0.1', port=4556):
        """Async variant of :meth:`TCPCLConnection.connect`. The interface is
        identical, except that this function is a coroutine and has the be
        awaited.
        """
        if not self.sock:
            return

        # Establish TCP connection
        await self.loop.sock_connect(
            self.sock, (host, port)
        )
        # Create stream reader and writer pair
        self.reader, self.writer = await asyncio.open_connection(
            sock=self.sock, loop=self.loop
        )
        logger.debug("TCPCL: Connected to %s:%d", host, port)

        # Send TCPCL contact header with own EID
        logger.debug("TCPCL: Send header for %r", self.eid)
        self.writer.write(serialize_tcpcl_contact_header(self.eid))
        await self.writer.drain()

        # Receive TCPCL contact header
        raw = await self.reader.read(1024)
        header = decode_tcpcl_contact_header(raw)
        logger.debug("TCPCL: Received header: %r", header)
        return header

    async def send(self, bundle):
        """Async variant of :meth:`TCPCLConnection.send`. The interface is
        identical, except that this function is a coroutine and has the be
        awaited.
        """
        if not self.sock:
            return

        self.writer.write(serialize_tcpcl_single_bundle_segment(bundle))
        return await self.writer.drain()

    async def __aexit__(self, *args):
        # Terminate TCPCL connection if established
        if self.writer:
            logger.debug("TCPCL: Terminate connection to %s:%d",
                         *self.writer.transport.get_extra_info('peername'))
            self.writer.write(b"\x50")
            await self.writer.drain()

            # Close TCP connection
            self.writer.close()

        self.writer = None
        self.reader = None
        self.sock = False


class TCPCLServer(object):
    """Simple asyncio-based TCPCL server

    By default this server does nothing with received messages. It is a simple
    data sink. Messages are handled by the :meth:`message_received` method
    which does nothing in the default implementation. Subclasses can overwrite
    this method to perform specific tasks on recevied TCPCL messages.

    This class can be used as async context manager to ensure the connections
    are closed correctly even in case of an error:

    .. code:: python

        import asyncio
        from pyd3tn.tcpcl import TCPCLServer

        async def listen():
            async with TCPCLServer('dtn:me', '127.0.0.1', 42420) as server:
                # do some other cool stuff here while the server is running in
                # the background ... or just wait for the server to finish
                await server.wait_closed()

        loop = asyncio.get_event_loop()
        loop.run_until_complete(listen())

    Args:
        eid (str): The endpoint identifier (EID) that is used for this DTN node
        host (str): Hostname / IP address the server should listening on
        port (int): TCP port number the server should listenin on
        backlog (int, optional): Number of concurrent TCP connections
        loop (asyncio.AbstractEventLoop, optional): Event loop that should be
            used. If not given the default :func:`asyncio.get_event_loop` will
            be used.
    """

    def __init__(self, eid, host, port, backlog=100, loop=None):
        self.eid = eid
        self.host = host
        self.port = port
        self.backlog = backlog
        self.server = None
        self.handlers = []
        self.loop = loop or asyncio.get_event_loop()

    async def start(self):
        """Start listening on the specified host and port"""
        if self.server:
            return

        self.server = await asyncio.start_server(
            self.client_connected, host=self.host, port=self.port,
            backlog=self.backlog, loop=self.loop
        )
        logger.debug("TCPCLServer: Listening on %s:%d", self.host, self.port)

    def close(self):
        """Terminate all open connections and close the listening socket"""
        if not self.server:
            return

        logger.debug("TCPCLServer: Closing")

        # Terminate all active handler tasks
        for task in self.handlers:
            task.cancel()

        # Close all connections and wait from them to be closed
        self.server.close()

    async def wait_closed(self):
        if not self.server:
            return

        await self.server.wait_closed()
        self.server = None

    async def client_connected(self, reader, writer):
        """Callback function executed whenever a new TCP connection is
        established.

        It creates a handler task for each new connection. Handler tasks are
        stored in the :attr:`handlers` list.

        Args:
            reader (asyncio.StreamReader): Reader for the underlying TCP
                connection
            writer (asyncio.StreamWriter): Writer for the underlying TCP
                connection
        """
        # We create an extra task for handling the connection because we want
        # to be able to cancel them. A call to "self.server.close()" would stop
        # this coroutine without any exception. Hence we have no possibility to
        # shutdown the TCPCL connection properly in this coroutine.
        task = self.loop.create_task(self.handle_connection(reader, writer))
        task.add_done_callback(lambda task: self.handlers.remove(task))
        self.handlers.append(task)

    async def handle_connection(self, reader, writer):
        """This method is executed for each new connection is a separate
        asyncio task. It sends and receives the TCPCL contact header and then
        receives messages and forwards them to the :meth:`received_message`
        method. In case of any error the TCPCL connection is terminated
        gracefully by sending `\x50` to the remote peer.

        Args:
            reader (asyncio.StreamReader): Reader for the underlying TCP
                connection
            writer (asyncio.StreamWriter): Writer for the underlying TCP
                connection
        """
        logger.debug("TCPCLServer: Connection from %s:%d",
                     *writer.transport.get_extra_info('peername'))
        try:
            writer.write(serialize_tcpcl_contact_header(self.eid))
            await writer.drain()

            # TCPCL header
            raw = await reader.read(1024)
            if not raw:
                logger.debug("TCPCLServer: Remote peer closed connection")
                return
            decode_tcpcl_contact_header(raw)

            # TCPCL messages
            while True:
                msg = await reader.read(1024)
                if not msg:
                    logger.debug("TCPCLServer: Remote peer closed connection")
                    return
                await self.received_message(msg, writer)
        except asyncio.CancelledError:
            pass
        # In case of any other exception, log the error and terminate the
        # connection
        except Exception as err:
            logger.exception(err)
        finally:
            logger.debug("TCPCLServer: Close connection from %s:%d",
                         *writer.transport.get_extra_info('peername'))

            # Gracefully terminate TCPCL session if TCP connection is not
            # already closing.
            if not writer.transport.is_closing():
                writer.write(b"\x50")
                await writer.drain()
                writer.close()

    async def received_message(self, msg, writer):
        """Message handler called whenever a message gets received. It does
        nothing but logging the message by default. Subclasses can overwrite
        this method to perform specific tasks.

        Args:
            msg (bytes): TCPCL message received
            writer (asyncio.StreamWriter): Writer for the TCP connection that
                can be used to react on the received message
        """
        logger.info("TCPCLServer: Received: %s", msg.hex())

    async def __aenter__(self):
        await self.start()
        return self

    async def __aexit__(self, *args):
        self.close()
        await self.wait_closed()



import enum
import time
import threading

from .helpers import UNIX_TO_DTN_OFFSET
from .sdnv import sdnv_encode


class RFC5050Flag(enum.IntFlag):
    """Default BPv6 flags"""
    IS_FRAGMENT = 0x00001
    ADMINISTRATIVE_RECORD = 0x00002
    MUST_NOT_BE_FRAGMENTED = 0x00004
    CUSTODY_TRANSFER_REQUESTED = 0x00008
    SINGLETON_ENDPOINT = 0x00010
    ACKNOWLEDGEMENT_REQUESTED = 0x00020

    # Priority
    NORMAL_PRIORITY = 0x00080
    EXPEDITED_PRIORITY = 0x00100

    # Reports
    REPORT_RECEPTION = 0x04000
    REPORT_CUSTODY_ACCEPTANCE = 0x08000
    REPORT_FORWARDING = 0x10000
    REPORT_DELIVERY = 0x20000
    REPORT_DELETION = 0x40000


RFC5050Flag.DEFAULT_OUTGOING = (
    RFC5050Flag.SINGLETON_ENDPOINT |
    RFC5050Flag.NORMAL_PRIORITY
)


class BlockType(enum.IntEnum):
    """(Extension) block types"""
    PAYLOAD = 1
    PREVIOUS_NODE = 7
    BUNDLE_AGE = 8
    HOP_COUNT = 9
    MAX = 255


class BlockFlag(enum.IntFlag):
    """(Extension) block flags"""
    NONE = 0x00
    MUST_BE_REPLICATED = 0x01
    REPORT_IF_UNPROC = 0x02
    DELETE_BUNDLE_IF_UNPROC = 0x04
    LAST_BLOCK = 0x08
    DISCARD_IF_UNPROC = 0x10
    FWD_UNPROC = 0x20
    HAS_EID_REF_FIELD = 0x40


NULL_EID = "dtn:none"


_th_local = threading.local()


def next_sequence_number():
    seqnum = _th_local.__dict__.get("seqnum", 0)
    _th_local.__dict__["seqnum"] = seqnum + 1
    return seqnum


def reset_sequence_number():
    _th_local.__dict__["seqnum"] = 0


def serialize_bundle6(source_eid, destination_eid, payload,
                      report_to_eid=NULL_EID, custodian_eid=NULL_EID,
                      creation_timestamp=None, sequence_number=None,
                      lifetime=300, flags=RFC5050Flag.DEFAULT_OUTGOING,
                      fragment_offset=None, total_adu_length=None):
    # RFC 5050 header: https://tools.ietf.org/html/rfc5050#section-4.5
    # Build part before "primary block length"
    header_part1 = bytearray()
    header_part1.append(0x06)
    header_part1 += sdnv_encode(flags)

    # Build part after "primary block length"
    header_part2 = bytearray()

    # NOTE: This does not do deduplication (which is optional) currently.
    dictionary = bytearray()
    cur_dict_offset = 0
    for eid in (destination_eid, source_eid, report_to_eid, custodian_eid):
        scheme, ssp = eid.encode("ascii").split(b":", 1)
        dictionary += scheme + b"\0"
        header_part2 += sdnv_encode(cur_dict_offset)
        cur_dict_offset += len(scheme) + 1
        dictionary += ssp + b"\0"
        header_part2 += sdnv_encode(cur_dict_offset)
        cur_dict_offset += len(ssp) + 1

    if creation_timestamp is None:
        creation_timestamp = int(time.time() - UNIX_TO_DTN_OFFSET)
    header_part2 += sdnv_encode(creation_timestamp)

    # If the sequence number is None, the last thread-local sequence number
    # will be used, incremented by one.
    if sequence_number is None:
        sequence_number = next_sequence_number()
    header_part2 += sdnv_encode(sequence_number)

    header_part2 += sdnv_encode(lifetime)

    header_part2 += sdnv_encode(len(dictionary))
    header_part2 += dictionary

    if fragment_offset is not None and total_adu_length is not None:
        assert (flags & RFC5050Flag.IS_FRAGMENT) != 0
        header_part2 += sdnv_encode(fragment_offset)
        header_part2 += sdnv_encode(total_adu_length)

    # Add the length of all remaining fields as primary block length
    header = header_part1 + sdnv_encode(len(header_part2)) + header_part2

    # Build payload block
    # https://tools.ietf.org/html/rfc5050#section-4.5.2
    payload_block = bytearray()
    payload_block.append(BlockType.PAYLOAD)
    payload_block += sdnv_encode(BlockFlag.LAST_BLOCK)
    # Block length is the length of the remaining part of the block (PL data)
    payload_block += sdnv_encode(len(payload))
    payload_block += payload

    # NOTE: This does _not_ support extension blocks currently
    return bytes(header + payload_block)



"""Generic CRC implementation with many pre-defined CRC models.

This module is the adopted Python implementation of the JavaScript CRC
implementation by Bastian Molkenthin:

    http://www.sunshine2k.de/coding/javascript/crc/crc_js.html

Copyright (c) 2015 Bastian Molkenthin

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

Usage:
    .. code:: python

        >>> from crc import crc32
        >>> checksum = crc32(b"Hello world")
        >>> hex(checksum)
        >>> '0x1b851995'

Tests:
    There are some very small unit tests that can be executed with ``pytest``:

    .. code:: bash

        pytest path/to/crc.py
"""

import array


def reflect(num, width):
    """Reverts bit order of the given number

    Args:
        num (int): Number that should be reflected
        width (int): Size of the number in bits
    """
    reflected = 0

    for i in range(width):
        if (num >> i) & 1 != 0:
            reflected |= 1 << (width - 1 - i)

    return reflected


def make_table(width):
    """Create static sized CRC lookup table and initialize it with ``0``.

    For 8, 16, 32, and 64 bit width :class:`array.array` instances are used.
    For all other widths it falls back to a generic Python :class:`list`.

    Args:
        width (int): Size of elements in bits
    """
    initializer = (0 for _ in range(256))

    if width == 8:
        return array.array('B', initializer)
    elif width == 16:
        return array.array('H', initializer)
    elif width == 32:
        return array.array('L', initializer)
    elif width == 64:
        return array.array('Q', initializer)
    else:
        # Fallback to a generic list
        return list(initializer)


class CRC(object):
    """Generic CRC model implemented with lookup tables.

    The model parameter can are the constructor parameters.

    Args:
        width (int): Number of bits of the polynomial
        polynomial (int): CRC polynomial
        initial_value (int): Initial value of the checksum
        final_xor_value (int): Value that will be XOR-ed with final checksum
        input_reflected (bool): True, if each input byte should be reflected
        result_reflected (bool): True, if the result should be reflected before
            the final XOR is applied

    Usage:
        .. code:: python

            from crc import CRC

            # Definition  of CRC-32 Ethernet. The crc module defines many
            # common models already.
            crc = CRC(32, 0x04c11db7, 0xffffffff, 0xffffffff, True, True)

            # You can call the model to calculate the CRC checksum of byte
            # string
            assert crc(b"Hello world!") == 0x1b851995
    """
    def __init__(self, width, polynomial, initial_value, final_xor_value,
                 input_reflected, result_reflected):
        self.width = width
        self.polynomial = polynomial
        self.initial_value = initial_value
        self.final_xor_value = final_xor_value
        self.input_reflected = input_reflected
        self.result_reflected = result_reflected

        # Initialize casting mask to keep the correct width for dynamic Python
        # integers
        self.cast_mask = int('1' * self.width, base=2)

        # Mask that can be applied to get the Most Significant Bit (MSB) if the
        # number with given width
        self.msb_mask = 0x01 << (self.width - 1)

        # The lookup tables get initialized lazzily. This ensures that only
        # tables are calculated that are actually needed.
        self.table = None
        self.reflected_table = None

    def __call__(self, value):
        """Compute the CRC checksum with respect to the model parameters by
        using a lookup table algorithm.

        Args:
            value (bytes): Input bytes that should be checked
        Returns:
            int - CRC checksum
        """
        # Use the reflection optimization if applicable
        if self.input_reflected and self.result_reflected:
            return self.fast_reflected(value)

        # Lazy initialization of the lookup table
        if self.table is None:
            self.table = self.calculate_crc_table()

        crc = self.initial_value

        for cur_byte in value:
            if self.input_reflected:
                cur_byte = reflect(cur_byte, 8)

            # Update the MSB of the CRC value with the next input byte
            crc = (crc ^ (cur_byte << (self.width - 8))) & self.cast_mask

            # This MSB byte value is the index into the lookup table
            index = (crc >> (self.width - 8)) & 0xff

            # Shift out the index
            crc = (crc << 8) & self.cast_mask

            # XOR-ing crc from the lookup table using the calculated index
            crc = crc ^ self.table[index]

        if self.result_reflected:
            crc = reflect(crc, self.width)

        # Final XBOR
        return crc ^ self.final_xor_value

    def fast_reflected(self, value):
        """If the input data and the result checksum are both reflected in the
        current model, an optimized algorithm can be used that reflects the
        looup table rather then the input data. This saves the reflection
        operation of the input data.
        """
        if not self.input_reflected or not self.result_reflected:
            raise ValueError("Input and result must be reflected")

        # Lazy initialization of the lookup table
        if self.reflected_table is None:
            self.reflected_table = self.calculate_crc_table_reflected()

        crc = self.initial_value

        for cur_byte in value:
            # The LSB of the XOR-red remainder and the next byte is the index
            # into the lookup table
            index = (crc & 0xff) ^ cur_byte

            # Shift out the index
            crc = (crc >> 8) & self.cast_mask

            # XOR-ing remainder from the loopup table
            crc = crc ^ self.reflected_table[index]

        # Final XBOR
        return crc ^ self.final_xor_value

    def calculate_crc_table(self):
        table = make_table(self.width)

        for divident in range(256):
            cur_byte = (divident << (self.width - 8)) & self.cast_mask

            for bit in range(8):
                if (cur_byte & self.msb_mask) != 0:
                    cur_byte <<= 1
                    cur_byte ^= self.polynomial
                else:
                    cur_byte <<= 1

            table[divident] = cur_byte & self.cast_mask

        return table

    def calculate_crc_table_reflected(self):
        table = make_table(self.width)

        for divident in range(256):
            reflected_divident = reflect(divident, 8)
            cur_byte = (
                (reflected_divident << (self.width - 8)) & self.cast_mask
            )

            for bit in range(8):
                if (cur_byte & self.msb_mask) != 0:
                    cur_byte <<= 1
                    cur_byte ^= self.polynomial
                else:
                    cur_byte <<= 1

            cur_byte = reflect(cur_byte, self.width)

            table[divident] = (cur_byte & self.cast_mask)

        return table


# Known CRC algorihtms
crc8 = CRC(8, 0x07, 0x00, 0x00, False, False)
crc8_sae_j1850 = CRC(8, 0x1d, 0xff, 0xff, False, False)
crc8_sae_j1850_zero = CRC(8, 0x1d, 0x00, 0x00, False, False)
crc8_8h2f = CRC(8, 0x2f, 0xff, 0xff, False, False)
crc8_cdma2000 = CRC(8, 0x9b, 0xff, 0x00, False, False)
crc8_darc = CRC(8, 0x39, 0x00, 0x00, True, True)
crc8_dvb_s2 = CRC(8, 0xd5, 0x00, 0x00, False, False)
crc8_ebu = CRC(8, 0x1d, 0xff, 0x00, True, True)
crc8_icode = CRC(8, 0x1d, 0xfd, 0x00, False, False)
crc8_itu = CRC(8, 0x07, 0x00, 0x55, False, False)
crc8_maxim = CRC(8, 0x31, 0x00, 0x00, True, True)
crc8_rohc = CRC(8, 0x07, 0xff, 0x00, True, True)
crc8_wcdma = CRC(8, 0x9b, 0x00, 0x00, True, True)

crc16_ccit_zero = CRC(16, 0x1021, 0x0000, 0x0000, False, False)
crc16_arc = CRC(16, 0x8005, 0x0000, 0x0000, True, True)
crc16_aug_ccitt = CRC(16, 0x1021, 0x1d0f, 0x0000, False, False)
crc16_buypass = CRC(16, 0x8005, 0x0000, 0x0000, False, False)
crc16_ccitt_false = CRC(16, 0x1021, 0xffff, 0x0000, False, False)
crc16_cdma2000 = CRC(16, 0xc867, 0xffff, 0x0000, False, False)
crc16_dds_110 = CRC(16, 0x8005, 0x800d, 0x0000, False, False)
crc16_dect_r = CRC(16, 0x0589, 0x0000, 0x0001, False, False)
crc16_dect_x = CRC(16, 0x0589, 0x0000, 0x0000, False, False)
crc16_dnp = CRC(16, 0x3d65, 0x0000, 0xffff, True, True)
crc16_en_13757 = CRC(16, 0x3d65, 0x0000, 0xffff, False, False)
crc16_genibus = CRC(16, 0x1021, 0xffff, 0xffff, False, False)
crc16_maxim = CRC(16, 0x8005, 0x0000, 0xffff, True, True)
crc16_mcrf4xx = CRC(16, 0x1021, 0xffff, 0x0000, True, True)
crc16_riello = CRC(16, 0x1021, 0xb2aa, 0x0000, True, True)
crc16_t10_dif = CRC(16, 0x8bb7, 0x0000, 0x0000, False, False)
crc16_teledisk = CRC(16, 0xa097, 0x0000, 0x0000, False, False)
crc16_tms37157 = CRC(16, 0x1021, 0x89ec, 0x0000, True, True)
crc16_usb = CRC(16, 0x8005, 0xffff, 0xffff, True, True)
crc16_a = CRC(16, 0x1021, 0xc6c6, 0x0000, True, True)
crc16_kermit = CRC(16, 0x1021, 0x0000, 0x0000, True, True)
crc16_modbus = CRC(16, 0x8005, 0xffff, 0x0000, True, True)
crc16_x25 = CRC(16, 0x1021, 0xffff, 0xffff, True, True)
crc16_xmodem = CRC(16, 0x1021, 0x0000, 0x0000, False, False)

crc32 = CRC(32, 0x04c11db7, 0xffffffff, 0xffffffff, True, True)
crc32_bzip2 = CRC(32, 0x04c11db7, 0xffffffff, 0xffffffff, False, False)
crc32_c = CRC(32, 0x1edc6f41, 0xffffffff, 0xffffffff, True, True)
crc32_d = CRC(32, 0xa833982b, 0xffffffff, 0xffffffff, True, True)
crc32_mpeg2 = CRC(32, 0x04c11db7, 0xffffffff, 0x00000000, False, False)
crc32_posix = CRC(32, 0x04c11db7, 0x00000000, 0xffffffff, False, False)
crc32_q = CRC(32, 0x814141ab, 0x00000000, 0x00000000, False, False)
crc32_jamcrc = CRC(32, 0x04c11db7, 0xffffffff, 0x00000000, True, True)
crc32_xfer = CRC(32, 0x000000af, 0x00000000, 0x00000000, False, False)


# ----------
# Unit tests
# ----------

def test_crc16_ccit_zero():
    assert crc16_ccit_zero(b'Hello world!') == 0x39db


def test_crc16_x25():
    assert crc16_x25(b'Hello world!') == 0x8edb


def test_crc32():
    assert crc32(b'Hello world!') == 0x1b851995
